package net.minestom.server.command.builder.arguments.minecraft;

import it.unimi.dsi.fastutil.chars.CharArrayList;
import it.unimi.dsi.fastutil.chars.CharList;
import net.minestom.server.command.CommandReader;
import net.minestom.server.command.builder.arguments.Argument;
import net.minestom.server.utils.time.TimeUnit;
import org.jetbrains.annotations.NotNull;

import java.time.Duration;
import java.time.temporal.TemporalUnit;

/**
 * Represents an argument giving a time (day/second/tick).
 * <p>
 * Example: 50d, 25s, 75t
 */
public class ArgumentTime extends Argument<Duration> {

    public static final int INVALID_TIME_FORMAT = -2;
    public static final int NO_NUMBER = -3;

    private static final CharList SUFFIXES = new CharArrayList(new char[]{'d', 's', 't'});

    public ArgumentTime(String id) {
        super(id);
    }

    @Override
    public @NotNull Result<Duration> parse(CommandReader reader) {
        String input = reader.readWord();
        final char lastChar = input.charAt(input.length() - 1);

        TemporalUnit timeUnit;
        if (Character.isDigit(lastChar))
            timeUnit = TimeUnit.SERVER_TICK;
        else if (SUFFIXES.contains(lastChar)) {
            input = input.substring(0, input.length() - 1);

            if (lastChar == 'd') {
                timeUnit = TimeUnit.DAY;
            } else if (lastChar == 's') {
                timeUnit = TimeUnit.SECOND;
            } else if (lastChar == 't') {
                timeUnit = TimeUnit.SERVER_TICK;
            } else {
                return Result.syntaxError("Time needs to have the unit d, s, t, or none", input, INVALID_TIME_FORMAT);
            }
        } else {
            return Result.syntaxError("Unknown time unit", input, INVALID_TIME_FORMAT);
        }
        try {
            // Check if value is a number
            final int time = Integer.parseInt(input);
            return Result.success(Duration.of(time, timeUnit));
        } catch (NumberFormatException e) {
            return Result.syntaxError("Time needs to be a number", input, NO_NUMBER);
        }
    }

    @Override
    public String parser() {
        return "minecraft:time";
    }

    @Override
    public String toString() {
        return String.format("Time<%s>", getId());
    }
}
